@@ -1,3 +1,5 @@
+from typing import Optional
+
 from rich.console import Console
 from rich.table import Table
 
@@ -6,13 +8,16 @@ from finetuner.model import list_model_classes
 console = Console()
 
 
-def print_model_table(model):
+def print_model_table(model, task: Optional[str] = None):
     """Prints a table of model descriptions.
 
     :param model: Module with model definitions
+    :param task: The fine-tuning task, should be one of `text-to-text`,
     """
-
-    table = Table(title='Finetuner backbones')
+    title = 'Finetuner backbones'
+    if task:
+        title += f': {task}'
+    table = Table(title=title)
     header = model.get_header()
     model_descriptors = set()
 
@@ -21,7 +26,10 @@ def print_model_table(model):
 
     for _, _model_class in list_model_classes().items():
         if _model_class.descriptor not in model_descriptors:
-            table.add_row(*model.get_row(_model_class))
+            row = model.get_row(_model_class)
+            if task and row[1] != task:
+                continue
+            table.add_row(*row)
             model_descriptors.add(_model_class.descriptor)
 
     console.print(table)
